@@ -25,6 +25,7 @@
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+#include <config.h>
 #include <stdio.h>
 #include <string.h>
 #include <openssl/crypto.h>
@@ -45,10 +46,12 @@
 static PKCS11_CTX *ctx;
 
 /** 
- * The PIN used for login. May be assigend by set_pin function, or by the
- * get_pin function (using an external UI). The memory for this PIN is always
- * owned internally, and may be freed as necessary. Before freeing, the PIN 
+ * The PIN used for login. Cache for the get_pin function.
+ * The memory for this PIN is always owned internally,
+ * and may be freed as necessary. Before freeing, the PIN 
  * must be whitened, to prevent security holes.
+ *
+ * length is always MAX_PIN_LENGTH and possibly not 0 terminated?
  */
 static char *pin = NULL;
 
@@ -56,9 +59,11 @@ static int verbose = 0;
 
 static char *module = NULL;
 
+static char *init_args = NULL;
+
 int set_module(const char *modulename)
 {
-	module = strdup(modulename);
+	module = modulename ? strdup(modulename) : NULL;
 	return 1;
 }
 
@@ -96,8 +101,10 @@ int inc_verbose(void)
 	return 1;
 }
 
-static char *get_pin(UI_METHOD * ui_method, void *callback_data, char *sc_pin,
-		     int maxlen)
+/* either get the pin code from the supplied callback data, or get the pin
+ * via asking our self. In both cases keep a copy of the pin code in the
+ * pin variable (strdup'ed copy). */
+static int get_pin(UI_METHOD * ui_method, void *callback_data)
 {
 	UI *ui;
 	struct {
@@ -105,11 +112,16 @@ static char *get_pin(UI_METHOD * ui_method, void *callback_data, char *sc_pin,
 		const char *prompt_info;
 	} *mycb = callback_data;
 
+	/* pin in the call back data, copy and use */
 	if (mycb->password) {
-		sc_pin = set_pin(mycb->password);
-		return sc_pin;
+		pin = (char *)calloc(MAX_PIN_LENGTH, sizeof(char));
+		if (!pin)
+			return 0;
+		strncpy(pin,mycb->password,MAX_PIN_LENGTH);
+		return 1;
 	}
 
+	/* call ui to ask for a pin */
 	ui = UI_new();
 	if (ui_method != NULL)
 		UI_set_method(ui, ui_method);
@@ -117,18 +129,24 @@ static char *get_pin(UI_METHOD * ui_method, void *callback_data, char *sc_pin,
 		UI_set_app_data(ui, callback_data);
 
 	if (!UI_add_input_string
-	    (ui, "PKCS#11 token PIN: ", 0, sc_pin, 1, maxlen)) {
+	    (ui, "PKCS#11 token PIN: ", 0, pin, 1, MAX_PIN_LENGTH)) {
 		fprintf(stderr, "UI_add_input_string failed\n");
 		UI_free(ui);
-		return NULL;
+		return 0;
 	}
 	if (UI_process(ui)) {
 		fprintf(stderr, "UI_process failed\n");
 		UI_free(ui);
-		return NULL;
+		return 0;
 	}
 	UI_free(ui);
-	return sc_pin;
+	return 1;
+}
+
+int set_init_args(const char *init_args_orig)
+{
+	init_args = init_args_orig ? strdup(init_args_orig) : NULL;
+	return 1;
 }
 
 int pkcs11_finish(ENGINE * engine)
@@ -139,7 +157,7 @@ int pkcs11_finish(ENGINE * engine)
 		ctx = NULL;
 	}
 	if (pin != NULL) {
-		OPENSSL_cleanse(pin, sizeof(pin));
+		OPENSSL_cleanse(pin, MAX_PIN_LENGTH);
 		free(pin);
 		pin = NULL;
 	}
@@ -152,6 +170,7 @@ int pkcs11_init(ENGINE * engine)
 		fprintf(stderr, "initializing engine\n");
 	}
 	ctx = PKCS11_CTX_new();
+        PKCS11_CTX_init_args(ctx, init_args);
 	if (PKCS11_CTX_load(ctx, module) < 0) {
 		fprintf(stderr, "unable to load module %s\n", module);
 		return 0;
@@ -162,6 +181,7 @@ int pkcs11_init(ENGINE * engine)
 int pkcs11_rsa_finish(RSA * rsa)
 {
 	if (pin) {
+		OPENSSL_cleanse(pin, MAX_PIN_LENGTH);
 		free(pin);
 		pin = NULL;
 	}
@@ -310,10 +330,19 @@ static int parse_slot_id_string(const char *slot_id, int *slot,
 
 	i = strspn(slot_id + 5, DIGITS);
 
-	if (slot_id[i + 5] == 0) {
+        /*
+         * changed 2008-01-28 9:31 GMT by SafeNet UK Ltd, akroehnert@safenet-inc.com
+         * changed comparision from == to >= as we need to check for any slot equal
+         * or greater zero
+         */
+	if (slot_id[i + 5] >= 0) {
 		*slot = n;
 		*id_len = 0;
-		return 1;
+	/*
+	 * changed 2008-01-28 9:31 GMT by SafeNet UK Ltd, akroehnert@safenet-inc.com
+      * if we jump out here we cant get the label or id parsed
+	 */
+	/*	return 1; */
 	}
 
 	if (slot_id[i + 5] != '-') {
@@ -341,7 +370,11 @@ static int parse_slot_id_string(const char *slot_id, int *slot,
 
 	/* ... or "label_" */
 	if (strncmp(slot_id + i, "label_", 6) == 0)
-		return (*label = strdup(slot_id + 6)) != NULL;
+		/*
+	         * changed 2008-01-28 9:31 CET by SafeNet UK Ltd, akroehnert@safenet-inc.com
+        	 * we have to chop off + i + 6 characters instead of just 6
+		 */
+		return (*label = strdup(slot_id + i + 6)) != NULL;
 
 	fprintf(stderr, "could not parse string!\n");
 	return 0;
@@ -524,6 +557,7 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 	if (s_slot_key_id && *s_slot_key_id) {
 		n = parse_slot_id_string(s_slot_key_id, &slot_nr,
 					 key_id, &key_id_len, &key_label);
+
 		if (!n) {
 			fprintf(stderr,
 				"supported formats: <id>, <slot>:<id>, id_<id>, slot_<slot>-id_<id>, label_<label>, slot_<slot>-label_<label>\n");
@@ -649,9 +683,9 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 		   not, allocate and obtain a new PIN. */
 		if (tok->secureLogin) {
 			/* Free the PIN if it has already been 
-			   assigned (i.e, via set_pin */
+			   assigned (i.e, cached by get_pin) */
 			if (pin != NULL) {
-				OPENSSL_cleanse(pin, strlen(pin));
+				OPENSSL_cleanse(pin, MAX_PIN_LENGTH);
 				free(pin);
 				pin = NULL;
 			}
@@ -660,14 +694,19 @@ static EVP_PKEY *pkcs11_load_key(ENGINE * e, const char *s_slot_key_id,
 			if (pin == NULL) {
 				fail("Could not allocate memory for PIN");
 			}
-			get_pin(ui_method, callback_data, pin, MAX_PIN_LENGTH);
+			if (!get_pin(ui_method, callback_data) ) {
+				OPENSSL_cleanse(pin, MAX_PIN_LENGTH);
+				free(pin);
+				pin = NULL;
+				fail("No pin code was entered");
+			}
 		}
 
 		/* Now login in with the (possibly NULL) pin */
 		if (PKCS11_login(slot, 0, pin)) {
 			/* Login failed, so free the PIN if present */
 			if (pin != NULL) {
-				OPENSSL_cleanse(pin, sizeof(pin));
+				OPENSSL_cleanse(pin, MAX_PIN_LENGTH);
 				free(pin);
 				pin = NULL;
 			}
